#!/usr/bin/perl

use strict;
use warnings;

use DBI;
use Getopt::Std;
use Term::ReadKey;
use HTML::Diff;

# Process command line arguments
my %opts;
getopts('sul:p:', \%opts);
my $version_one = shift or die "Usage: $0 [-su] [-l <lang>] [-p <mysql_password>] <version_one> <version_two> <version_three>\n";
my $version_two = shift or die "Usage: $0 [-su] [-l <lang>] [-p <mysql_password>] <version_one> <version_two> <version_three>\n";
my $version_three = shift or die "Usage: $0 [-su] [-l <lang>] [-p <mysql_password>] <version_one> <version_two> <version_three>\n";

my $lang = $opts{l} || 'en';

# Connect to database
my ($data_source, $db_user);
if ($opts{s}) {
  # ssh -L 6666:127.0.0.1:3306 gibus@pi.serverside.fr
	$data_source = 'dbi:mysql:dbname=pi_tpdb;host=127.0.0.1;port=6666';
	$db_user = 'pi_tpdb';
}
else {
	$data_source = 'dbi:mysql:dbname=tpdb';
	$db_user = 'gibus';
}

# Get mysql password
my $db_passwd;
if (defined $opts{p}) {
    $db_passwd = $opts{p};
}
else {
    print "Enter mysql passwd for $db_user:\n";
    ReadMode 'noecho';
    $db_passwd = ReadLine 0;
    chomp $db_passwd;
    ReadMode 'normal';
}

my $dbh = DBI->connect($data_source, $db_user, $db_passwd) or die $DBI::errstr;

# Get 'one' chunks
my $sql_one = "SELECT * FROM tp_text WHERE version='$version_one' AND lang='$lang'";
my $sth_one = $dbh->prepare($sql_one);
$sth_one->execute;
my $one = $sth_one->fetchall_hashref( [ qw(amended number) ] );

# Get 'two' chunks
my $sql_two = "SELECT * FROM tp_text WHERE version='$version_two' AND lang='$lang'";
my $sth_two = $dbh->prepare($sql_two);
$sth_two->execute;
my $two = $sth_two->fetchall_hashref( [ qw(amended number) ] );

# Get 'three' chunks
my $sql_three = "SELECT * FROM tp_text WHERE version='$version_three' AND lang='$lang'";
my $sth_three = $dbh->prepare($sql_three);
$sth_three->execute;
my $three = $sth_three->fetchall_hashref( [ qw(amended number) ] );

# Compute diff chunks
my %diff;
# Iterate through 'two' chunks
foreach my $amended (keys %{$two}) {
    foreach my $number (keys %{$two->{$amended}}) {
        my $two_chunk = $two->{$amended}->{$number}->{text};
        # Find related 'one' chunk if any
        my $one_chunk;
        my $num;
        if (exists $one->{$amended}) {
            # 'one' chunk with same number
            if (exists $one->{$amended}->{$number}) {
                $one_chunk = $one->{$amended}->{$number}->{text};
                $num = $number;
                $diff{$amended}->{$num}->{number_one} = $number;
            }
            # 'one' chunk with added paragraphs on the same level
            elsif (exists $one->{$amended}->{$number.'1'}) {
                $one_chunk = $one->{$amended}->{$number.'1'}->{text};
                $num = $number.'1';
                $diff{$amended}->{$num}->{number_one} = $number.'1';
            }
            # 'one' chunk with added paragraphs on the previous level
            elsif (exists $one->{$amended}->{substr($number, 0, length($number)-1).'1'.substr($number, -1, 1)}) {
                $one_chunk = $one->{$amended}->{substr($number, 0, length($number)-1).'1'.substr($number, -1, 1)}->{text};
                $num = substr($number, 0, length($number)-1).'1'.substr($number, -1, 1);
                $diff{$amended}->{$num}->{number_one} = substr($number, 0, length($number)-1).'1'.substr($number, -1, 1);
            }
            # 'one' chunk with only one paragraphs on the same level
            elsif (substr($number, -1, 1) eq '1' and exists $one->{$amended}->{substr($number, 0, length($number)-1)}) {
                $one_chunk = $one->{$amended}->{substr($number, 0, length($number)-1)}->{text};
                $num = $number;
                $diff{$amended}->{$num}->{number_one} = substr($number, 0, length($number)-1);
            }
            # 'one' chunk with only one paragraphs on the previous level
            elsif (substr($number, -2, 1) eq '1' and exists $one->{$amended}->{substr($number, 0, length($number)-2).substr($number, length($number)-1,1)}) {
                $one_chunk = $one->{$amended}->{substr($number, 0, length($number)-2).substr($number, length($number)-1,1)}->{text};
                $num = $number;
                $diff{$amended}->{$num}->{number_one} = substr($number, 0, length($number)-2).substr($number, length($number)-1,1);
            }
        }

        # No 'one' chunk, 'two' chunk is new
        unless (defined $one_chunk) {
            $one_chunk = '';
            $num = $number;
            $diff{$amended}->{$num}->{number_one} = undef;
        }

        # Find related 'three' chunk if any
        my $three_chunk;
        if (exists $three->{$amended}) {
            # 'three' chunk with same number
            if (exists $three->{$amended}->{$number}) {
                $three_chunk = $three->{$amended}->{$number}->{text};
                $diff{$amended}->{$num}->{number_three} = $number;
            }
            # 'three' chunk with added paragraphs on the same level
            elsif (exists $three->{$amended}->{$number.'1'}) {
                $three_chunk = $three->{$amended}->{$number.'1'}->{text};
                if ($num ne $number.'1') {
                    $diff{$amended}->{$number.'1'}->{number_one} = delete $diff{$amended}->{$num}->{number_one};
                    delete $diff{$amended}->{$num};
                    $num = $number.'1';
                }
                $diff{$amended}->{$num}->{number_three} = $number.'1';
            }
            # 'three' chunk with added paragraphs on the previous level
            elsif (exists $three->{$amended}->{substr($number, 0, length($number)-1).'1'.substr($number, -1, 1)}) {
                $three_chunk = $three->{$amended}->{substr($number, 0, length($number)-1).'1'.substr($number, -1, 1)}->{text};
                if ($num ne substr($number, 0, length($number)-1).'1'.substr($number, -1, 1)) {
                    $diff{$amended}->{substr($number, 0, length($number)-1).'1'.substr($number, -1, 1)}->{number_one} = delete $diff{$amended}->{$num}->{number_one};
                    delete $diff{$amended}->{$num};
                    $num = substr($number, 0, length($number)-1).'1'.substr($number, -1, 1);
                }
                $diff{$amended}->{$num}->{number_three} = substr($number, 0, length($number)-1).'1'.substr($number, -1, 1);
            }
            # 'three' chunk with only one paragraphs on the same level
            elsif (substr($number, -1, 1) eq '1' and exists $three->{$amended}->{substr($number, 0, length($number)-1)}) {
                $three_chunk = $three->{$amended}->{substr($number, 0, length($number)-1)}->{text};
                $diff{$amended}->{$num}->{number_three} = substr($number, 0, length($number)-1);
            }
            # 'three' chunk with only one paragraphs on the previous level
            elsif (substr($number, -2, 1) eq '1' and exists $three->{$amended}->{substr($number, 0, length($number)-2).substr($number, length($number)-1,1)}) {
                $three_chunk = $three->{$amended}->{substr($number, 0, length($number)-2).substr($number, length($number)-1,1)}->{text};
                $diff{$amended}->{$num}->{number_three} = substr($number, 0, length($number)-2).substr($number, length($number)-1,1);
            }
        }

        # No 'three' chunk, 'two' chunk is new
        unless (defined $three_chunk) {
            $three_chunk = '';
            $diff{$amended}->{$num}->{number_three} = undef;
        }

        # Compute diffs
        $diff{$amended}->{$num}->{text_one} = $one_chunk;
        if ($one_chunk) {
            $diff{$amended}->{$num}->{text_two} = diff($one_chunk, $two_chunk);
            $diff{$amended}->{$num}->{text_three} = diff($one_chunk, $three_chunk);
            delete $one->{$amended}->{$diff{$amended}->{$num}->{number_one}};
        }
        else {
            if ($two_chunk) {
                $diff{$amended}->{$num}->{text_two} = diff($one_chunk, $two_chunk);
            }
            else {
                $diff{$amended}->{$num}->{text_two} = $two_chunk;
            }

            if ($three_chunk) {
                $diff{$amended}->{$num}->{text_three} = diff($one_chunk, $three_chunk);
            }
            else {
                $diff{$amended}->{$num}->{text_three} = $three_chunk;
            }
        }
        delete $three->{$amended}->{$diff{$amended}->{$num}->{number_three}} if $three_chunk;

        # Build result
        $diff{$amended}->{$num}->{version_one} = $version_one;
        $diff{$amended}->{$num}->{version_two} = $version_two;
        $diff{$amended}->{$num}->{version_three} = $version_three;
        $diff{$amended}->{$num}->{nature} = $two->{$amended}->{$number}->{nature};
        $diff{$amended}->{$num}->{amending} = $two->{$amended}->{$number}->{amending};
        $diff{$amended}->{$num}->{amended} = $two->{$amended}->{$number}->{amended};
        $diff{$amended}->{$num}->{number_two} = $number;
        $diff{$amended}->{$num}->{lang} = $two->{$amended}->{$number}->{lang};
    }
}

# Iterate through 'three' chunks that have not been deleted
foreach my $amended (keys %{$three}) {
    foreach my $number (keys %{$three->{$amended}}) {
        my $three_chunk = $three->{$amended}->{$number}->{text};
        # Find related 'one' chunk if any
        my $one_chunk;
        my $num;
        if (exists $one->{$amended}) {
            # 'one' chunk with same number
            if (exists $one->{$amended}->{$number}) {
                $one_chunk = $one->{$amended}->{$number}->{text};
                $num = $number;
                $diff{$amended}->{$num}->{number_one} = $number;
            }
            # 'one' chunk with added paragraphs on the same level
            elsif (exists $one->{$amended}->{$number.'1'}) {
                $one_chunk = $one->{$amended}->{$number.'1'}->{text};
                $num = $number.'1';
                $diff{$amended}->{$num}->{number_one} = $number.'1';
            }
            # 'one' chunk with added paragraphs on the previous level
            elsif (exists $one->{$amended}->{substr($number, 0, length($number)-1).'1'.substr($number, -1, 1)}) {
                $one_chunk = $one->{$amended}->{substr($number, 0, length($number)-1).'1'.substr($number, -1, 1)}->{text};
                $num = substr($number, 0, length($number)-1).'1'.substr($number, -1, 1);
                $diff{$amended}->{$num}->{number_one} = substr($number, 0, length($number)-1).'1'.substr($number, -1, 1);
            }
            # 'one' chunk with only one paragraphs on the same level
            elsif (substr($number, -1, 1) eq '1' and exists $one->{$amended}->{substr($number, 0, length($number)-1)}) {
                $one_chunk = $one->{$amended}->{substr($number, 0, length($number)-1)}->{text};
                $num = $number;
                $diff{$amended}->{$num}->{number_one} = substr($number, 0, length($number)-1);
            }
            # 'one' chunk with only one paragraphs on the previous level
            elsif (substr($number, -2, 1) eq '1' and exists $one->{$amended}->{substr($number, 0, length($number)-2).substr($number, length($number)-1,1)}) {
                $one_chunk = $one->{$amended}->{substr($number, 0, length($number)-2).substr($number, length($number)-1,1)}->{text};
                $num = $number;
                $diff{$amended}->{$num}->{number_one} = substr($number, 0, length($number)-2).substr($number, length($number)-1,1);
            }
        }

        # No 'one' chunk, 'three' chunk is new
        unless (defined $one_chunk) {
            $one_chunk = '';
            $num = $number;
            $diff{$amended}->{$num}->{number_one} = undef;
        }

        # No 'two' chunk, 'three' chunk is new
        my $two_chunk = '';
        $diff{$amended}->{$num}->{number_two} = undef;

        # Compute diffs
        $diff{$amended}->{$num}->{text_one} = $one_chunk;
        if ($one_chunk) {
            $diff{$amended}->{$num}->{text_two} = diff($one_chunk, $two_chunk);
            delete $one->{$amended}->{$diff{$amended}->{$num}->{number_one}};
        }
        else {
            $diff{$amended}->{$num}->{text_two} = $two_chunk;
        }
        $diff{$amended}->{$num}->{text_three} = diff($one_chunk, $three_chunk);

        # Build result
        $diff{$amended}->{$num}->{version_one} = $version_one;
        $diff{$amended}->{$num}->{version_two} = $version_two;
        $diff{$amended}->{$num}->{version_three} = $version_three;
        $diff{$amended}->{$num}->{nature} = $three->{$amended}->{$number}->{nature};
        $diff{$amended}->{$num}->{amending} = $three->{$amended}->{$number}->{amending};
        $diff{$amended}->{$num}->{amended} = $three->{$amended}->{$number}->{amended};
        $diff{$amended}->{$num}->{number_three} = $number;
        $diff{$amended}->{$num}->{lang} = $three->{$amended}->{$number}->{lang};
    }
}

# Iterate through 'one' chunks that have not been deleted
foreach my $amended (keys %{$one}) {
    foreach my $number (keys %{$one->{$amended}}) {
        $diff{$amended}->{$number}->{version_one} = $version_one;
        $diff{$amended}->{$number}->{version_two} = $version_two;
        $diff{$amended}->{$number}->{version_three} = $version_three;
        $diff{$amended}->{$number}->{nature} = $one->{$amended}->{$number}->{nature};
        $diff{$amended}->{$number}->{amending} = $one->{$amended}->{$number}->{amending};
        $diff{$amended}->{$number}->{amended} = $one->{$amended}->{$number}->{amended};
        $diff{$amended}->{$number}->{number_three} = undef;
        $diff{$amended}->{$number}->{number_two} = undef;
        $diff{$amended}->{$number}->{number_one} = $number;
        $diff{$amended}->{$number}->{lang} = $one->{$amended}->{$number}->{lang};
        $diff{$amended}->{$number}->{text_one} = $one->{$amended}->{$number}->{text};
        $diff{$amended}->{$number}->{text_two} = "<s>$one->{$amended}->{$number}->{text}</s>";
        $diff{$amended}->{$number}->{text_three} = "<s>$one->{$amended}->{$number}->{text}</s>";

        unless ($diff{$amended}->{$number}->{amending}) {
            $diff{$amended}->{$number}->{amending} = $diff{$amended}->{$number}->{amended} < 4 ? 1 : 2;
        }
    }
}

# Insert diff in mysql
my ($ins, $upd, $err) = (0, 0, 0);
foreach my $amended (keys %diff) {
    foreach my $number (keys %{$diff{$amended}}) {
        # Check if text is already known in database
        my $sql_check = "SELECT number FROM tp_three_texts WHERE version_one='$version_one' AND version_two='$version_two' AND version_three='$version_three' AND nature='$diff{$amended}->{$number}->{nature}' AND amending='$diff{$amended}->{$number}->{amending}' AND amended='$diff{$amended}->{$number}->{amended}' AND number='$number' AND lang='$lang'";
        (my $check) = $dbh->selectrow_array($sql_check);
        unless ($check) {
            my $sql_insert = "INSERT INTO tp_three_texts VALUES ('$diff{$amended}->{$number}->{version_one}','$diff{$amended}->{$number}->{version_two}','$diff{$amended}->{$number}->{version_three}','$diff{$amended}->{$number}->{nature}','$diff{$amended}->{$number}->{amending}','$diff{$amended}->{$number}->{amended}','$number',".($diff{$amended}->{$number}->{number_one} ? "'$diff{$amended}->{$number}->{number_one}'" : "NULL").",".($diff{$amended}->{$number}->{number_two} ? "'$diff{$amended}->{$number}->{number_two}'" : "NULL").",".($diff{$amended}->{$number}->{number_three} ? "'$diff{$amended}->{$number}->{number_three}'" : "NULL").",'$lang',".$dbh->quote($diff{$amended}->{$number}->{text_one}).",".$dbh->quote($diff{$amended}->{$number}->{text_two}).",".$dbh->quote($diff{$amended}->{$number}->{text_three}).")";
            $dbh->do($sql_insert);
            warn "$diff{$amended}->{$number}->{version_one} $diff{$amended}->{$number}->{version_two} $diff{$amended}->{$number}->{version_three} $diff{$amended}->{$number}->{nature} $diff{$amended}->{$number}->{amending} $diff{$amended}->{$number}->{amended} $number $lang inserted\n";
            $ins++;
        }
        elsif ($opts{u}) {
            my $sql_update = "UPDATE tp_three_texts SET text_one=".$dbh->quote($diff{$amended}->{$number}->{text_one}).",text_two=".$dbh->quote($diff{$amended}->{$number}->{text_two}).",text_three=".$dbh->quote($diff{$amended}->{$number}->{text_three})." WHERE version_one='$version_one' AND version_two='$version_two' AND version_three='$version_three' AND nature='$diff{$amended}->{$number}->{nature}' AND amending='$diff{$amended}->{$number}->{amending}' AND amended='$diff{$amended}->{$number}->{amended}' AND number='$number' AND lang='$lang'";
            $dbh->do($sql_update);
            warn "$diff{$amended}->{$number}->{version_one} $diff{$amended}->{$number}->{version_two} $diff{$amended}->{$number}->{version_three} $diff{$amended}->{$number}->{nature} $diff{$amended}->{$number}->{amending} $diff{$amended}->{$number}->{amended} $number $lang updated\n";
            $upd++;
        }
        else {
  	        warn "Error: cannot insert $diff{$amended}->{$number}->{version_one} $diff{$amended}->{$number}->{version_two} $diff{$amended}->{$number}->{version_three} $diff{$amended}->{$number}->{nature} $diff{$amended}->{$number}->{amending} $diff{$amended}->{$number}->{amended} $number $lang\n";
            $err++;
        }
    }
}
print "Inserted: $ins, updated: $upd, errors: $err\n";

# Compute diff
sub diff {
    my $from = shift;
    my $to = shift;
    my @diff = @{html_word_diff($from, $to)};
    my $diff = '';
    foreach my $chunk (@diff) {
        # Unchanged
        if ($chunk->[0] eq 'u') {
            $diff .= $chunk->[1];
        }
        # Deleted
        elsif ($chunk->[0] eq '-') {
            $diff .= "<s>".$chunk->[1]."</s>";
        }
        # Added or updated
        else {
            $diff .= "'''''".$chunk->[2]."'''''";
        }
    }
    return $diff;
}
