#!/usr/bin/perl
use strict;
use warnings;

use File::Basename;
use DBI;
use Getopt::Std;
use Term::ReadKey;
use open ':utf8';

# Process optional command line arguments
my %opts;
getopts('sul:p:', \%opts);

# Open input file
my $file = shift or die "Usage: $0 [-su] [-l <lang>] [-p <mysql_password>] <input_file>\n";
open IN, $file or die "Cannot open $file: $!\n";

# Guess version, amending, amended and lang from file name
my ($version, $amending, $amended, $lang) = (basename($file, '.txt') =~ /^([a-z])(\d)(\d)([a-z][a-z])$/o);

# Connect to database
my ($data_source, $db_user);
if ($opts{s}) {
  # ssh -L 6666:127.0.0.1:3306 gibus@pi.serverside.fr
	#$data_source = 'dbi:mysql:dbname=pi_tpdb;host=127.0.0.1;port=6666;mysql_enable_utf8=1';
	$data_source = 'dbi:mysql:dbname=pi_tpdb;host=127.0.0.1;port=6666';
	$db_user = 'pi_tpdb';
}
else {
	$data_source = 'dbi:mysql:dbname=tpdb';
	$db_user = 'gibus';
}

# Get mysql password
my $db_passwd;
if (defined $opts{p}) {
    $db_passwd = $opts{p};
}
else {
    print "Enter mysql passwd for $db_user:\n";
    ReadMode 'noecho';
    $db_passwd = ReadLine 0;
    chomp $db_passwd;
    ReadMode 'normal';
}

my $dbh = DBI->connect($data_source, $db_user, $db_passwd) or die $DBI::errstr;

# Parse input file
my ($ins, $upd, $err) = (0, 0, 0);
while (<IN>) {
  if (s/^(\d) ([^ ]+) //) {
    db_insert($1, $2, $_);
  }
}
print "Inserted: $ins, updated: $upd, errors: $err\n";

sub db_insert {
  my ($nature, $number, $text) = @_;
  chomp $text;
  # Check if text is already known in database
  my $sql_check = "SELECT number FROM tp_text WHERE version='$version' AND nature='$nature' AND amending='$amending' AND amended='$amended' AND number='$number' AND lang='$lang'";
  (my $check) = $dbh->selectrow_array($sql_check);
  unless ($check) {
    my $sql_insert = "INSERT INTO tp_text VALUES ('$version', '$nature', '$amending', '$amended', '$number', '$lang', ".$dbh->quote($text).")";
    $dbh->do($sql_insert);
    warn "$version $nature $amending $amended $number $lang inserted\n";
    $ins++;
  }
  elsif ($opts{u}) {
    my $sql_update = "UPDATE tp_text SET text=".$dbh->quote($text)." WHERE version='$version' AND nature='$nature' AND amending='$amending' AND amended='$amended' AND number='$number' AND lang='$lang'";
    $dbh->do($sql_update);
    warn "$version $nature $amending $amended $number $lang updated\n";
    $upd++;
  }
  else {
  	warn "Error: cannot insert $version $nature $amending $amended $number $lang\n";
    $err++;
  }
}
